/************************************************************************
	printr.pro


	Example background print utility
	
	printr <filename>  
	 loads the utility if it is not already resident 
	 starts printing file name, 
		or adds it to the queue of files to be printed
	printr /pause
	  pauses printing
        printr /resume
	  restarts printing
	printr /cancel
	  stops all printing and unloads the resident process
        printr
          lists the name of file in progress + files in queue


	Copyright (C) 1990, Prolog Development Center
************************************************************************/
heap = $500

include "resident.pre"

CONSTANTS
  %% try_ems_Swap	= 1	
     progname 		= "printr"
    	
     pause_printing 	= 3
     resume_printing	= 4
     display_queue  	= 5
     print_slice 	= 1 	

     form_feed		= '\012'
DOMAINS
    file = pfile

DATABASE
    printing_file 	
    file_queue(string)

PREDICATES
    process_comline(string)
    valid_file(string)
    status_request
    do_switch(string)
    trim(string,string)
    open_as_Input(String)
    best_Swap
    printer_not_busy(integer)
    unbuffered_print(char)	

GOAL	
	comline(Stuff),
	upper_lower(Stuff,Commands),
	process_comline(Commands).
	
CLAUSES
%
% timer event handler: 
%   installed so that it is called approx once per timer tick
%
        
    timer(print_slice) :-
	not(eof(pfile)),
	!,
	%% see if the printer is busy yet...
        printer_not_busy(5),
        readChar(Char),
	unbuffered_print(Char),
        timer(print_slice).	   % keep going
    timer(print_slice) :-
	closefile(pfile),
	unbuffered_print(form_feed),
        retract(file_queue(_)),
    	file_queue(File),!, 	    % still one left
    	open_as_input(File).
    timer(print_slice) :-           % else done, remove self
        removeResident(progname).   % could just use "exit"

%% check the printer status a few times in succession,
%%   it is *briefly* busy immediately after recieving a character
%%   which is a different sense of busy than printer buffer full 
    printer_not_busy(_) :- 
        bios($17,reg($0200,0,0,$0000,0,0,0,0),reg(AX,_,_,_,_,_,_,_)),
        bitand(AX, $8000, NotBusyBit), 
	NotBusyBit = $8000,!.
    printer_not_busy(C) :- C2=C-1, C2>0, printer_not_busy(C2).

%% send character to the printer:
%%  originally we used a re-directed "write", but Prolog buffers 
%%  printer output, which means the printer-busy condition
%%  could occur as the buffer is written, defeating the
%%  purpose of the printer-busy test, and leaving a too-long
%%  print-slice interval (resulting in choppy DOS response)

    unbuffered_print(C) :-	
	bios($17,reg(C,0,0,0,0,0,0,0),_).

%
% Signal handlers... 
%   signal is raised by non-resident copy in response 
%    to command-line switch, eg,
%    >printr /pause
%
    signal(pause_printing) :-!,
    	setStatusTimer(print_slice,off).	

    signal(resume_printing) :-!,
    	setStatusTimer(print_slice,on).
	
    signal(data_received) :-
    	rdDataResident(progname, FileName),
    	assertz(file_queue(FileName)).

    	
    signal(display_queue) :-
	cursorsync,
	write("\n\nPrinting or waiting to print:\n"),
    	file_queue(File), 
    	write(File), nl,
    	fail.
    signal(display_queue):- 
	write("==="),
	cursorstate(on).
    
ifdef try_ems_swap
	best_Swap :- 
		enableSwap(swapToEms, "","", Status),
		Status=0,!,
		write(progname, " installed for EMS swap\n").
enddef
/******************* not enough time in slice for this! ************
	best_Swap :- 
		enableSwap(swapToFile, "printr.swp","", Status),
		Status=0,!,
		write(progname, " installed for disk swap").
********************************************************************/
	best_Swap:- write(progname, " installed with no swapping\n").


    open_As_input(File) :-
    	openread(pfile, File),
    	filemode(pfile,0),  % don't muck with CRs
	readdevice(pfile).

    process_comline(Command) :-
    	not(fronttoken(Command,_,_)),!,
    	status_request.
    process_comline(Comline) :-
    	fronttoken(Comline , "/", Rest),!,
	fronttoken(Rest, Switch,_),
	do_switch(Switch).
    process_comline(MaybeFilename) :-
    	trim(MaybeFileName, FileName),
    	trap(existfile(Filename),_,fail),!,
    	valid_file(FileName).
    process_comline(Stuff) :-
    	write("file ",Stuff," not found!").
    	

    status_request :-
	cursorstate(on),
    	signalResident(progname, display_queue),!.
    status_request :-
    	write("Print utility is not active").
    	
    valid_file(Filename) :-
    	enableResident(progname, 80),!,  % buffer[80] to receive filenames
	best_Swap,
	assertz(file_queue(FileName)),
	installSignal(pause_printing,0),
	installSignal(resume_printing,0),
	installSignal(data_received, needsLib), 
	QueueFlags = needsLib + needsDos + stopTimers,
	installSignal(display_queue, QueueFlags),
	open_as_input(FileName),
	write("Printing: ", Filename),
	installTimer(print_slice, 6, needsDos_cantReenter),
	stayResident.

    %% failed enableResident, so we can presume that
    %%  it is already resident & printing... 
    %%  so we just send it the file name	

    valid_file(Filename) :-  
	write("adding ", Filename, " to queue"),
    	wrDataResident(progname, FileName).   % pass filename to resident copy

    %% here, a second, non-resident instance 
    %% signals the resident instance of the program

    do_switch(pause) :-
    	signalResident(progname, pause_printing),!.  
    do_switch(resume) :-
    	signalResident(progname, resume_printing),!.
    do_switch(cancel) :- 
	removeResident(progname),!,
	write("\nprinting cancelled.").
    do_switch(Switch) :-
    	Switch<>"pause",
    	Switch<>"resume",
	Switch<>"cancel",!,
       	write("invalid option:  /pause  /resume  /cancel  or <filename>").
    do_switch(_) :-		% signal failed on valid switch
    	write("Print utility is not active").
    	    		

    %% trim the blanks from each side of string 
    trim(A,B) :-
    	frontchar(A, 32, A2),
    	!,trim(A2, B).
    trim(A, B) :-
    	concat(A2, " ", A),
    	!,trim(A2, B).	
    trim(A,A).
    
    
    	
      	
    		
