/*
=====================================================================
	GRAPH.PRO
	
	Relative scaled graphic system. Supports window relative
	coordinate system and scaled object display.
	
	Copyright (C) 1990 Prolog Development Center
=====================================================================
*/

/*
---------------------------------------------------------------------
	GRAPHIC SCALE HANDLING
---------------------------------------------------------------------
*/

DATABASE - scaling

	scale_(Integer,Integer,Integer,Integer,Integer)
	active_scale_(Integer)

CLAUSES

   % DEFAULT SCALE VALUES

	active_scale_(0).
	scale_(0,0,1000,0,1000).      % Default is the window scale


   % CREATE A NEW SCALE
	
	make_scale(Number,XMin,XMax,YMin,YMax) :-
		retractall(scale_(Number,_,_,_,_)),
		assert(scale_(Number,XMin,XMax,YMin,YMax)),
		retractall(active_scale_(_)), !,
		assert(active_scale_(Number)).


   % SHIFT TO ANOTHER SCALE

	shift_scale(Number) :- free(Number),
		active_scale_(Number), !.
	shift_scale(Number) :- bound(Number),
		retractall(active_scale_(_)), !,
		assert(active_scale_(Number)).


   % DELETE A SCALE

	remove_scale(Number) :-
		retractall(scale_(Number,_,_,_,_)), !.

/*
---------------------------------------------------------------------
	RELATIVE SPACE TRANSLATION PREDICATES
---------------------------------------------------------------------
*/

   % TRANSLATE COORDINATES TO AND FROM RELATIVE SPACE

	window_pix(XRel,YRel,XPix,YPix,Rad) :-
		getviewsettings(Left,Top,Right,Bot,_),
		Width = Right - Left, 
		Height = Bot - Top,
		active_scale_(Active),
		scale_(Active,XMin,XMax,YMin,YMax), !,
		XRange = XMax - XMin, 
		YRange = YMax - YMin,
		XPix = (Width / XRange) * (XRel - XMin),
		YPix = Height - (Height / YRange) * (YRel - YMin),
		Rad = (Height / YRange) * YRel.

	window_pix(XRel,YRel,XRad,YRad) :-
		getviewsettings(Left,Top,Right,Bot,_),
		Width = Right - Left, 
		Height = Bot - Top,
		active_scale_(Active),
		scale_(Active,XMin,XMax,YMin,YMax), !,
		XRange = XMax - XMin, 
		YRange = YMax - YMin,
		XRad = (Width / XRange) * XRel,
		YRad = (Height / YRange) * YRel.


    % TRANSLATE A POINT LIST TO AND FROM RELATIVE SPACE

	window_list([],[]) :- !.
	window_list([XRef,YRef|Tail],[XPix,YPix|Back]) :-
		window_pix(XRef,YRef,XPix,YPix,_),
		window_list(Tail,Back).


/*
---------------------------------------------------------------------
	GRAPHIC WINDOW HANDLING
---------------------------------------------------------------------
*/

DATABASE - graphic_windows

	window_(Integer,Integer,Integer,Integer,Integer,Integer,Integer)
	inside_(Integer,Integer,Integer,Integer,Integer)
	active_window_(Integer)

CLAUSES

   % MAKE A WINDOW TITLE BOX IF ONE IS REQUIRED
	
	make_title(Top,Left,Bottom,Right,"",Attr,Tip) :- !,
		split_attr(Attr,Foreground,_),
		setcolor(Foreground),
		rectangle(Left,Top,Right,Bottom), 
		Tip = Top + 1.
	make_title(Top,Left,Bottom,Right,Text,Attr,Tip) :-
		settextjustify(1,1), 
		textheight(Text,Height),
		Bot = Top + Height + 2, 
		Tip = Bot + 1,
		Middle = Top + (Height / 2) + 2,
		Center = ((Right - Left) / 2) + Left,
		split_attr(Attr,Foreground,Background),
		setfillstyle(1,Background), 
		setcolor(Foreground),
		bar(Left,Top,Right,Bot),
		rectangle(Left,Top,Right,Bot),
		outtextxy(Center,Middle,Text),
		rectangle(Left,Bot,Right,Bottom).


   % CREATE AND PLACE A NEW WINDOW		

	make_Gwindow(Number,WAttr,FAttr,Title,Top,Left,Height,Width) :-
		Bottom = Top + Height, 
		Right = Left + Width,
		getmaxx(MaxX), 
		getmaxy(MaxY),
		setviewport(0,0,MaxX,MaxY,clip),
		make_title(Top,Left,Bottom,Right,Title,FAttr,Bot),
		L = Left + 1, 
		R = Right - 1, 
		B = Bottom - 1,
		split_attr(WAttr,Foreground,Background),
		setfillstyle(1,Background), 
		setcolor(Foreground),
		bar(L,Bot,R,B), 
		setviewport(L,Bot,R,B,clip),
		retractall(window_(Number,_,_,_,_,_,_)),
		retractall(active_window_(_)), !,
		asserta(window_(Number,Background,L,Top,R,B,Bot)),
		assert(active_window_(Number)).


   % REMOVE A GIVEN WINDOW

	remove_Gwindow(Number) :-
		window_(Number,_,Left,Top,Right,Bottom,_), !,
		L = Left - 1, 
		R = Right + 1,
		T = Top, 
		B = Bottom + 1,
		retractall(window_(Number,_,_,_,_,_,_)),
		retractall(inside_(Number,_,_,_,_)), !,
		setviewport(L,T,R,B,clip), clearviewport.


   % CREATE AN INTERNAL WINDOW

	make_inside(Left,Top,Right,Bottom) :-     % Scaled coordinates
		window_pix(Left,Top,X1,Y1,_),
		window_pix(Right,Bottom,X2,Y2,_),
		getviewsettings(L,T,_,_,_),
		VLeft = L + X1 + 1, 
		VRight = L + X2 - 1,
		VBottom = T + Y1 + 1, 
		VTop = T + Y2 - 1,
		setviewport(VLeft,VTop,VRight,VBottom,clip),
		active_window_(Window), !,
		retractall(inside_(Window,_,_,_,_)),
		assert(inside_(Window,VLeft,VTop,VRight,VBottom)).


   % CLEAR THE CURRENT WINDOW WITH BACKGROUND WIPE
		
	clear_Gwindow :- active_window_(Number),
		window_(Number,Background,_,_,_,_,_), !,
		getviewsettings(L,T,R,B,_),
		Width = R - L + 1, 
		Height = B - T + 1,
		setfillstyle(1,Background),
		bar(0,0,Width,Height).
		

   % SHIFT TO A GIVEN WINDOW

	shift_Gwindow(Number) :- free(Number),
		active_window_(Number), !.
	shift_Gwindow(Number) :- bound(Number),
		inside_(Number,Left,Top,Right,Bottom), !,
		setviewport(Left,Top,Right,Bottom,clip),
		retractall(active_window_(_)), !,
		assert(active_window_(Number)).
	shift_Gwindow(Number) :- bound(Number),
		window_(Number,_,Left,_,Right,Bottom,T), !,
		L = Left + 1, 
		R = Right - 1, 
		B = Bottom - 1,
		setviewport(L,T,R,B,clip),
		retractall(active_window_(_)),
		assert(active_window_(Number)).

   % SPLIT ATTRIBUTE VALUES INTO FOREGROUND AND BACKGROUND COLORS

	split_attr(Attr,Foreground,Background) :-
		bitand(15,Attr,LowByte),
		Foreground = LowByte,
		bitand(240,Attr,HighByte), !,
		Background = HighByte / 16.

/*
---------------------------------------------------------------------
	GRAPHIC TEXT MESSAGES
---------------------------------------------------------------------
*/

   % DISPLAY A LIST OF TEXT IN THE CURRNET WINDOW

	show_text([Head|List]) :- textheight(Head,Height),
		clear_Gwindow, 
		settextjustify(1,0),
		type_message(10,Height,[Head|List]).
   
 
	type_message(Old,Increment,[Head|Tail]) :-
   		getviewsettings(Left,_,Right,_,_),
   		Middle = (Right - Left) / 2,
   		outtextxy(Middle,Old,Head),
	   	New = Old + Increment, !,
   		type_message(New,Increment,Tail).
	type_message(_,_,_).

/*
---------------------------------------------------------------------
	RELATIVE SPACE DRAWING PREDICATES
---------------------------------------------------------------------
*/

   % PLOT A SINGLE POINT OF GIVEN SIZE AND TYPE IN RELATIVE SPACE

	draw_point(XRel,YRel,_,plot_pix) :-
		window_pix(XRel,YRel,XPix,YPix,_), !,
		getcolor(Color), putpixel(XPix,YPix,Color).
	draw_point(XRel,YRel,Radius,plot_circle) :-
		window_pix(XRel,YRel,XPix,YPix,_), !,
		ellipse(XPix,YPix,0,360,Radius,Radius).
	draw_point(XRel,YRel,Radius,plot_dot) :-
		window_pix(XRel,YRel,XPix,YPix,_), !,
		fillellipse(XPix,YPix,Radius,Radius).
	draw_point(XRel,YRel,Size,plot_box) :-
		window_pix(XRel,YRel,XPix,YPix,_),
		Left = XPix - Size, Right = XPix + Size,
		Top = YPix - Size, Bottom = YPix + Size, !,
		rectangle(Left,Top,Right,Bottom).
	draw_point(XRel,YRel,Size,plot_fill) :-
		window_pix(XRel,YRel,XPix,YPix,_),
		Left = XPix - Size, Right = XPix + Size,
		Top = YPix - Size, Bottom = YPix + Size, !,
		bar(Left,Top,Right,Bottom).
	draw_point(XRel,YRel,Size,plot_plus) :-
		window_pix(XRel,YRel,XPix,YPix,_),
		Left = XPix - Size, Right = XPix + Size,
		Top = YPix - Size, Bottom = YPix + Size, !,
		line(Left,YPix,Right,YPix),
		line(XPix,Bottom,XPix,Top).
	draw_point(XRel,YRel,Size,plot_cross) :-
		window_pix(XRel,YRel,XPix,YPix,_),
		Left = XPix - Size, Right = XPix + Size,
		Top = YPix - Size, Bottom = YPix + Size, !,
		line(Left,Top,Right,Bottom),
		line(Left,Bottom,Right,Top).


   % DRAW A LINE IN RELATIVE SPACE

	draw_line(X1Rel,Y1Rel,X2Rel,Y2Rel) :-
		window_pix(X1Rel,Y1Rel,X1Pix,Y1Pix,_),
		window_pix(X2Rel,Y2Rel,X2Pix,Y2Pix,_), !,
		line(X1Pix,Y1Pix,X2Pix,Y2Pix).


   % DRAW A RECTANGULAR BAR IN RELATIVE SPACE	
	
	draw_bar(X1Rel,Y1Rel,X2Rel,Y2Rel) :-
		getfillsettings(_,Color), Color = 0,
		window_pix(X1Rel,Y1Rel,X1Pix,Y1Pix,_),
		window_pix(X2Rel,Y2Rel,X2Pix,Y2Pix,_),
		rectangle(X1Pix,Y1Pix,X2Pix,Y2Pix), !.
	draw_bar(X1Rel,Y1Rel,X2Rel,Y2Rel) :-
		window_pix(X1Rel,Y1Rel,X1Pix,Y1Pix,_),
		window_pix(X2Rel,Y2Rel,X2Pix,Y2Pix,_), !,
		bar(X1Pix,Y1Pix,X2Pix,Y2Pix),
		rectangle(X1Pix,Y1Pix,X2Pix,Y2Pix).


   % DRAW A 3D BAR IN RELATIVE SPACE

	draw_bar3d(X1Rel,Y1Rel,X2Rel,Y2Rel,TopFlag) :-
		window_pix(X1Rel,Y1Rel,X1Pix,Y1Pix,_),
		window_pix(X2Rel,Y2Rel,X2Pix,Y2Pix,_),
		Depth = (X2Pix - X1Pix) / 4,
		bar3d(X1Pix,Y1Pix,X2Pix,Y2Pix,Depth,TopFlag).
	

   % DRAW AN ARC IN RELATIVE SPACE
	
	draw_arc(XRel,YRel,Start,End,XRadius,YRadius) :-
		window_pix(XRel,YRel,XPix,YPix,_),
		window_pix(XRadius,YRadius,XRad,_,YRad),
		ellipse(XPix,YPix,Start,End,XRad,Yrad).


   % DRAW AN ELLIPSE IN RELATIVE SPACE		

	draw_ellipse(XRel,YRel,XRadius,YRadius) :-
		window_pix(XRel,YRel,XPix,YPix,_),
		window_pix(XRadius,YRadius,XRad,_,YRad), !,
		fillellipse(XPix,YPix,XRad,YRad).


   % DRAW A PIE SLICE IN RELATIVE SPACE

	draw_pie(XRel,YRel,Start,End,Radius) :-	
		window_pix(XRel,YRel,XPix,YPix,_),
		window_pix(0,Radius,_,_,Rad), !,
		pieslice(XPix,YPix,Start,End,Rad).

	draw_pie(XRel,YRel,Start,End,XRadius,YRadius) :-	
		window_pix(XRel,YRel,XPix,YPix,_),
		window_pix(XRadius,YRadius,XRad,YRad), !,
		pieslicexy(XPix,YPix,Start,End,XRad,YRad).


   % DRAW A POLYGON IN RELATIVE SPACE

	draw_poly(PointList) :-
		getfillsettings(_,Color), Color = 0, !,
		window_list(PointList,PolyList),
		drawpoly(PolyList).
	draw_poly(PointList) :-
		window_list(PointList,PolyList),
		fillpoly(PolyList).


   % DRAW FORMATTED AND ADJUSTED TEXT

	draw_text(XRel,YRel,Form,Text,no_flag) :-
		window_pix(XRel,YRel,XPix,YPix,_),
		XPix >= 0, YPix >= 0, !,
		format(String,Form,Text),
		outtextxy(XPix,YPix,String).
	draw_text(XRel,YRel,Form,Text,vert_flag) :- !,
		window_pix(XRel,YRel,XPix,YPix,_),
		format(String,Form,Text),
		textwidth(String,Width),
		X = XPix - (Width / 2) - 4,
		outtextxy(X,YPix,String).
	draw_text(XRel,YRel,Form,Text,horz_flag) :- !,
		window_pix(XRel,YRel,XPix,YPix,_),
		format(String,Form,Text),
		textheight(String,Height),
		Y = YPix + (Height / 2) + 2,
		outtextxy(XPix,Y,String).
	draw_text(_,_,_,_,_).


	draw_set([style(Color,Fill,Style)|Tail]) :-
		setcolor(Color), setfillstyle(Style,Fill), !,
		draw_set(Tail).
	draw_set([draw_point(XPos,YPos,Rad,Type)|Tail]) :-
		draw_point(XPos,YPos,Rad,Type), !,
		draw_set(Tail).
	draw_set([draw_line(X1,Y1,X2,Y2)|Tail]) :-
		draw_line(X1,Y1,X2,Y2), !,
		draw_set(Tail).
	draw_set([draw_bar(X1,Y1,X2,Y2)|Tail]) :-
		draw_bar(X1,Y1,X2,Y2), !,
		draw_set(Tail).
	draw_set([draw_bar3d(X1,Y1,X2,Y2,Flag)|Tail]) :-
		draw_bar3d(X1,Y1,X2,Y2,Flag), !,
		draw_set(Tail).
	draw_set([draw_arc(XPos,YPos,Start,End,XRad,YRad)|Tail]) :-
		draw_arc(XPos,YPos,Start,End,XRad,YRad), !,
		draw_set(Tail).
	draw_set([draw_ellipse(XPos,YPos,XRad,YRad)|Tail]) :-
		draw_ellipse(XPos,YPos,XRad,YRad), !,
		draw_set(Tail).
	draw_set([draw_pie(XPos,YPos,Start,End,Rad)|Tail]) :-
		draw_pie(XPos,YPos,Start,End,Rad), !,
		draw_set(Tail).
	draw_set([draw_poly(PointList)|Tail]) :-
		draw_poly(PointList), !,
		draw_set(Tail).
	draw_set([draw_text(XPos,YPos,Form,Text,Flag)|Tail]) :-
		draw_text(XPos,YPos,Form,Text,Flag), !,
		draw_set(Tail).
	draw_set([drawing_(Drawing)|Tail]) :-
		draw_set(Drawing), !, draw_set(Tail).
	draw_set(_).

