/*
===========================================================================  
 	XPRINTER.PRO

  	Read the file XPRINTER.PRO (this file) - convert everything 
  	to uppercase and send it to a LaserJet compatible printer

  	Currently set up for XON/XOFF flow control...see your
  	 printer manual to verify what its factory settings are

	Copyright (C) 1990 Prolog Development Center
===========================================================================
*/

check_determ
project "xprinter"

CONSTANTS  
  port=1
  retry_count=10
  retry_wait=50
  
GLOBAL DOMAINS
  file = sp
  
include "SERIAL\\COMGLOBS.PRE"

PREDICATES
  safe_print
  print
  process_until_eof
  fileopen(file,string)
  transmit(char)
  trans_ch_delay(char,integer)
  wait_until_empty_output_buffer(integer,integer)

CLAUSES

/****************************************************************
	Open a file for reading - if possible
*****************************************************************/

fileopen(Sp,FileName):-
	trap(openread(Sp,FileName),_,fail),  % emulate old-style
	!, readdevice(Sp).
fileopen(_,FileName) :-
	write("\nImpossible to open ",FileName),
	fail.


/****************************************************************
	Read a character from current readdevice
	convert it to uppercase and print it
	Continue this process until eof or a time out event
*****************************************************************/

  process_until_eof:-
  	eof(sp),
  	transmit('\012'),!. % Send Form Feed 
  process_until_eof:-
	readchar(CH), 
	upper_lower(UpperCH,CH),
	transmit(UpperCH), 
	process_until_eof.

  transmit('\010'):-!,	    % map "nl" to crlf	
  	trans_ch_delay('\013',retry_count),
  	trans_ch_delay('\010',retry_count).
  transmit(CH)    :-trans_ch_delay(CH,retry_count).

  trans_ch_delay(CH,_):-
  	txCh_RS232(port,CH),!. 
  trans_ch_delay(CH,Retry):-
  	Retry>0, !,
  	R2=Retry-1,
  	sleep(retry_wait),
  	trans_ch_delay(CH,R2).
  trans_ch_delay(_,_) :-
	closefile(sp), closeRS232(port),
	write("\nPrinter is not ready, Program aborted"), 
	fail.


/****************************************************************
	Wait until output buffer is empty
	 time out if nothing happening for too long...
*****************************************************************/

  wait_until_empty_output_buffer(_,_):-
	queuesize_RS232(port,_,Queue), 
	Queue=0,!.
  
  wait_until_empty_output_buffer(_, LastSize):-
  	sleep(retry_wait),			
  	queuesize_RS232(port,_ ,OutQueue),	% as long as something 
  	OutQueue<LastSize,!, 			% happening, keep waiting 
  	wait_until_empty_output_buffer(retry_count, OutQueue).
  	
  wait_until_empty_output_buffer(Retry,Last):- 	 % if still nothing happens
	Retry>0,!, 		      	 	 % start retry countdown  	
	R2=Retry-1,
	wait_until_empty_output_buffer(R2,Last).

  wait_until_empty_output_buffer(_,_):-
  	write("\nTimed out").


/****************************************************************/


  print :-			
	InputBufSize	=	1,	% Size of input buffer 
	OutputBufSize	=	4000,	% Size of output buffer 
	BaudRate	=	b9600,	% baud
	Parity		=	parity_none,
	WordLength	=	data8,	% data bits	,
	StopBits	=	stop1,	% One stop bits	
  	Protocol	=	xon_xoff,% ADD dsr_dtr if your printer
  					 % cable supports it and your
  					 % printer is set to use it
  	openRs232(port, InputBufSize, OutputBufSize, BaudRate, Parity,
		  WordLength, StopBits, Protocol),
	fileopen(sp,"XPRINTER.PRO"), 
	!,
	process_until_eof, 
	closefile(sp),
	write("\nPrint Succeded\nWaiting to empty buffer..."),
	wait_until_empty_output_buffer(retry_count, OutputBufsize),
	closeRS232(port).

  safe_print :-			% safe use of the serial routines means	
  	trap(print,_,fail),!.	% insuring that that no matter what
  safe_print :-			% happens, the interrupt vectors are
  	closeRS232(port),!.	% properly restored!
  safe_print.
   

GOAL
	 safe_print
