/*
===========================================================================
	XXMODEM.PRO

	
   	Xmodem File Transfer
  
  	This demo implements the "checksum" version of XMODEM...
  	It may be easily extended to the 1024 block size & CRC versions
  	by detecting the CRC startup character ('C'), increasing the buffer
  	size, the block-size read, and by substituting a 2-byte CRC value
  	for the single-byte checksum.    
  
  	Various timing parameters are defined in the constants at the start
  	of the program and may be adjusted to create a more "relaxed" XMODEM
  	as may be required for use with time-sharing information services.

	Protocol signal bytes
   	ACK		    6		acknowledgment
   	NAK		   21		Negative acknowledgment
   	SOH		    1		Start of data character
   	EOT		    4		End of transmisssion

	Packet format:
       	1                  2                  3-130             131
   	Packet No,   Complement to Packet No,   Data (128) ...,   Check sum

	Protocol:   

   	TRANSMITTER				RECEIVER
   	===========                             ========
						Send NAK (repeat until SOH)
   	Send SOH
   	Send packet
   					Everything is ok:
   						Send NAK
   						Everything is not ok:
   						Send ACK

   	This will be repeated until transmission is accepted.
   	When all packges are transmitted the transmitter will:
   	
   	Send EOT

	Copyright (C) 1990 Prolog Development Center
============================================================================
*/

check_determ
project "xxmodem"

CONSTANTS
 ack=6
 nak=21
 soh=1
 eot=4 

 binary_mode = 0
 
 buffer_size = 256
 block_size  = 128
 packet_size = 131
 fill_byte   = 26
 
 port_no=2
  
% TIMING CONSTANTS: all times in centiseconds
   
% initial nak-ing to start transmitter (up to 9 nak 10 sec apart)
  initial_nak_retries = 9
  initial_soh_wait    = 1000
  first_nak_wait      = 1000

% consecutive retries on erroneous block 
  max_naks	     = 9

% can wait 20 seconds for a new block to start
  wait_for_block     = 2000 

% or once started, can wait 20 seconds to finish
  wait_in_block	     = 2000	
   
% sender wait for ack 30 seconds
  ack_wait  	     = 3000

GLOBAL DOMAINS
  FILE = infile ; outfile

include "serial\\comglobs.pre"
include "ui\\status.pro"
include "ui\\filename.pro"
include "ui\\menu.pro"

DOMAINS
  receipt = reply_was(char);
  		timed_out

  xmit_status = open(string,string); 
  		close_report;
  		progress(integer, real);
  		error

DATABASE - xmodem
 determ total_errors(INTEGER)

PREDICATES
  send_file(string)
  send_bytes(real,integer)
  send_block(integer,integer, integer,string)
  send_case(receipt, integer, real, integer, string)  
  wait_first_nak
    
  receive_file(string)
  receive_blocks(char,integer,integer)
  wait_new_header(integer,integer,real)
  valid_start(char)
        
  nak_until_initial_soh(integer)
  nak_until_case(receipt,integer)
  
  wait_for_reply(receipt, integer)
  wait_for(integer,integer)
  wait_aux(integer,real)
      
  complement(char,char)
  fill_if_less_than_128(integer,char)	
  lessor_of(real,real,real)
  checksum(string,integer,char)
  checksum_aux(integer,integer,integer,char,char)

  message(string)
  report(xmit_status)
  report_on(xmit_status)
 
  
ifndef in_xcomm
  port(integer)  
  safedemo
  demo
  decide(integer)

GOAL   safedemo.

CLAUSES

safedemo :-
	trap(demo, _, fail),!.		% all serial code should be
safedemo :-				% trapped to be sure that
	port(Port),			% ports get closed, OTHERWISE
	closeRS232(Port),		% Interrupts may not be restored!!!
	fail.
safedemo :- 
	existwindow(2),
	removewindow(2,1),
	fail.
safedemo :-
	removewindow(1,1).
	
demo :-	
	textmode(Row,Col),
	WinLast = Row - 1,
	makewindow(1, 26, 26, " Xmodem File Transfer ", 0, 0, WinLast, Col),
	port(PortNo), 				% COM1 
	InputBufSize	=	buffer_size,	% Size of input buffer 
	OutputBufSize	=	buffer_size,	% Size of output buffer 
	BaudRate	=	b2400,		% 9600 bits per second	
	Parity		=	parity_none,	% No parity		
	WordLength	=	data8,	 	% Eight data bits	
	StopBits	=	stop1,	 	% One stop bits	
  	Protocol	= 	0, 	% Xon/Xoff can not be used 
  	openRs232(PortNo, InputBufSize, OutputBufSize, BaudRate, Parity,
  		  WordLength, StopBits, Protocol),
	repeat,
	menu(10,20,64,23,
	     ["Transmit File",
  	      "Receive  File",
     	      "Quit"],
              "Choose an option",0,Choice),
	decide(Choice),
	Choice=3,!,
	closeRS232(PortNo),
	removewindow(1,1).

CLAUSES
  /* Transmit a file using Packaging */
  decide(1):-
  	makestatus(7,"Enter name of file to transmit"),
	readfilename(12,20,7,7,dat,"", Filename, existing_file),
	report(open("Sending",Filename)),
	send_file(Filename),
	report(close_report),
	!.
  
  /* Receive a file using Packaging */
  decide(2):-
  	makestatus(7, "Enter name of file to receive"),
	readfilename(13,20,7,7,dat,"",Filename,new_file),
	report(open("Receiving",Filename)),  	
	receive_file(Filename), 
	report(close_report),
	!.
  
  /* Quit */
  decide(3).

enddef

/***********************************************************************
   Receive file predicates
 ***********************************************************************/

receive_file(Filename) :-
	port(Port),
	delInbuf_RS232(Port),		% be sure no leftovers or noise
	nak_until_initial_soh(initial_nak_retries),  % mating dance
	openwrite(outfile, Filename),		     % really ready now	
	filemode(outfile, binary_mode),		     %  so open file	
	writedevice(outfile),
	receive_blocks(soh,max_naks,1),		     %  tro until eot		
	writedevice(screen),
	closefile(outfile),!.			     %  or aborted...	
receive_file(_).

receive_blocks(eot,_, _) :-!,
	port(Port),
	txCh_RS232(Port,ack).	% end of transmission
receive_blocks(soh,0,_ ) :- 	% nak limit exceeded for this block!
	!,
	message("Receive aborted...Too many errors").
receive_blocks(soh, _, Expecting) :-
	report(progress(Expecting, 0)),
	wait_for(packet_size, wait_in_block),	% dont even check until
	port(Port),
	rxCH_RS232(Port, BlockNo),		% all there
	BlockNo=Expecting mod 256,		% such that all fails here 
	rxCH_RS232(Port,Complement),		% are either 
	complement(Expecting,Complement),	%  block took too long
	rxBin_RS232(Port,Block,block_size),	%   or had bad data
	rxCH_RS232(Port,CheckSum),
	checksum(Block,block_size,Checksum),	% last check OK
	txCh_RS232(Port,ack),			% Ack the xmitter
	writeblock(block_size,Block),		%  write to file
	!,					
	Next=Expecting+1,
	marktime(wait_for_block,TimeAllowed),	% expect next block
	wait_new_header(max_naks, Next, TimeAllowed).
receive_blocks(_, Retry, Expect) :-		% failed checks or timeout
	port(Port),
	delInbuf_RS232(Port),			%  empty in case timeout
	txCH_RS232(Port,nak),			% nak & retry 
	R2=Retry-1,
	report(error),
	marktime(wait_for_block, TimeAllowed),
	wait_new_header(R2, Expect,TimeAllowed).
	
wait_new_header(_, Expect,_) :-			% discard anything
	port(Port),
	rxCh_RS232(Port,Header),		% until valid header
	valid_start(Header),
	!, receive_blocks(Header, max_naks, Expect).	
wait_new_header(Retry,Expect,EndTime) :- 	% or if time expires, fail.
	not(timeout(EndTime)),
	wait_new_header(Retry,Expect,EndTime).


valid_start(soh). 
valid_start(eot).

/***************************************************************************
   send file predicates
****************************************************************************/
send_file(Filename) :-
	trap(openread(infile, Filename),_,fail),
	filemode(infile,binary_mode),
	filepos(infile,0,2),		% end of file
	filepos(infile,FileSize,0),	%  = Filesize
	filepos(infile,0,0),		% reposition at start
	readdevice(infile),
	wait_first_nak,
	send_bytes(FileSize, 1),	% send Size bytes, starting block no
	readdevice(keyboard),
	closefile(infile),!.
send_file(File) :-
	write("Cant read ",File),!.
	
send_bytes(0, _):- 
	port(Port),
	txCh_RS232(Port, eot),!.	% no more bytes to send, notify done
send_bytes(Remaining, BlockNo) :-			
	report(progress(BlockNo, Remaining)),
	lessor_of(Remaining, block_size, DataSize),	% block or leftover
	readblock(DataSize, Block),				
	send_block(max_naks, DataSize, BlockNo, Block), % (re)send until ack
	Next=BlockNo+1,
	Rem2=Remaining-DataSize,
	!, send_bytes(Rem2, Next).
send_bytes(_,_). 			% catch send_block fail

send_block(0 , _,_, _) :- !, 
	message("Send: Retry count exceeded...too many errors!"),
	fail.
send_block( Retry , Size, BlockNo, Block) :-
	complement(BlockNo, CBkNo),		%
	port(Port),
	txCH_RS232(Port, soh),			% checking for naks here
	txCH_RS232(Port, BlockNo),		% is silly...its all buffered
	txCH_RS232(Port, CBkNo),		%  so check after sent
	txBin_RS232(Port, Block,Size),		%  if early nak, delete 
	checksum(Block,Size, CheckSum),		%  outbuffer
	fill_if_less_than_128(Size, CheckSum),	% fills and sends, or sends
	wait_for_reply(Reply, ack_wait ),
	send_case(Reply, Retry, Size, BlockNo, Block).	

send_case(timed_out,_,_,_,_):-			% cant get ack, give up!
	message("Send: Timeout waiting for ack"), 
	!, fail.
send_case(reply_was(ack),_,_,_ ,_) :-!.		% ok!
send_case(reply_was(nak),Retry, Size, No, Block) :- % receive asked for retry
	Retry2= Retry-1,
	report(error),
	!, send_block(Retry2, Size, No, Block).	
%send_case(reply_was(_) , R, S, P, B) :-  % received something else ...
%	--> invalid reply..        .	  % XMODEM not well defined here...


%% they also serve who only sit and wait...
%%
wait_for_reply(reply_was(Byte),Wait) :-
	wait_for(1, Wait),
	port(Port),
	rxCH_RS232(Port,Byte),
	!.
wait_for_reply(timed_out,_).

wait_for(Incoming, Interval) :-
	marktime(Interval, EndTime),
	wait_aux(Incoming, EndTime).

wait_aux(Expecting,_) :-
	port(Port),
	queuesize_RS232(Port, Inqueue, _),
	Inqueue>=Expecting,!.	
wait_aux(Expecting, Time) :-
	not(timeout(Time)),
	wait_aux(Expecting,Time).
	  
%
% fill out the last block to 128 sent, then send checksum
%  fill_byte constant specifies fill character (normally CTRL-Z)
%  herein lies the source of XMODEM's habit of rounding file sizes up
%  to the next multiple of 128 bytes... a vestigial bit of CPM...

fill_if_less_than_128(128, CheckSum) :-!,
	port(Port),
	txCH_RS232(Port,CheckSum).
fill_if_less_than_128(LastSent, CheckSum) :-
	port(Port),
	txCH_RS232(Port,fill_byte),
	Last2=LastSent+1,
	Check2=(CheckSum+fill_byte) mod 256,
	!, fill_if_less_than_128(Last2,  Check2).
	

% Compute/test 1s complement of a byte
%
complement(Byte, CompByte) :- 
	bitxor(Byte,255,Comp), 
	CompByte=Comp.
                  
%
% Compute/test the 1 byte checksum on the binary block
%
checksum(Ptr, Size, CheckSum) :-
	ptr_dword(Ptr,Seg,Off),
	checksum_aux( Size, Seg, Off, 0, CheckSum).

checksum_aux(0, _, _, Sum, Sum):-!. 
checksum_aux(N, Seg, Off, Accum, Total) :-
	membyte(Seg,Off,Byte),
	A2=(Accum+Byte) mod 256,
	N2=N-1,
	Off2=Off+1,
	checksum_aux(N2, Seg, Off2, A2, Total).


nak_until_initial_soh(0) :-!,
	message("No transmit...Receiver timeout!"),
	fail.	
nak_until_initial_soh(RetryNo) :-   
	port(Port),
	txCH_RS232(Port,nak),			% nak and...
	wait_for_reply(Reply,initial_soh_wait), % wait up to 10 seconds
	nak_until_case(Reply,RetryNo).		%  for a reply
	
nak_until_case(reply_was(soh),_) :- !.	
nak_until_case(reply_was(_),Retry) :-
	message("reply but not soh!"),
	report(error),
	port(Port),
	delInbuf_RS232(Port),            % out of whack, flush & re-sync
	R2 = Retry -1,	
	!,nak_until_initial_soh(R2).
nak_until_case(timed_out, Retry) :-	 % eg, 1 nak every 10 seconds until...
	R2=Retry-1,
	report(error),
	nak_until_initial_soh(R2).
	
wait_first_nak :-
	message("Waiting for nak from receiver..."),
	wait_for_reply(Reply,first_nak_wait),
	Reply=reply_Was(nak),!.
wait_first_nak :- wait_first_nak.

lessor_of(A,B,A) :- A<B,!.
lessor_of(_, B, B).

message(String) :- 
	writedevice(SaveD),
	writedevice(screen),
	
	write(String),
	nl,
	writedevice(SaveD).
	
report(open(Mode,Filename)) :-!,
	format(Title," %: % ", Mode, Filename),
	makewindow(2, 7,7, Title, 10, 4, 4, 55),
	retractall(total_errors(_)), 
	asserta(total_errors(0)).
	
report(close_report) :- !,
	cursor(0,0), write("Press a key to continue"), readchar(_),
	removewindow(2,1).

report(X) :-
	shiftwindow(Was),
	writedevice(Saved),
	writedevice(screen),
	gotowindow(2),
	report_on(X),
	gotowindow(Was),
	writedevice(Saved).

report_on(progress(Block, Remaining)) :-
	cursor(1,3),
	writef("Block %-6   Remaining %-6  Errors (none)", Block, Remaining).
		
report_on(error) :-	
	retract(total_errors(T)),!,
	T2=T+1,
	asserta(total_errors(T2)),
	cursor(1, 36),
	writef("Errors %-6", T2).

ifndef in_xcomm
 port(port_no).
enddef

